import os
INFLUENCE_XINETICS = os.path.dirname(os.path.abspath(__file__)) + "/data/influence_dm5v2.fits"
INFLUENCE_BMC_KILO = os.path.dirname(os.path.abspath(__file__)) + "/data/influence_BMC_kiloDM_300micron_res10_spline.fits"
INFLUENCE_BMC_2K = os.path.dirname(os.path.abspath(__file__)) + "/data/influence_BMC_2kDM_400micron_res10.fits"

